import os
import sys
import argparse
from prodir import (
    create_tree_structurer,
    get_structure,
    create_structure_from_file
)

def main():
    # Create the main parser
    parser = argparse.ArgumentParser(
        prog='prodir',  # Set program name to prodir
        description='Directory structure tool: Display and create directory structures'
    )
    
    # Create subparsers for different commands
    subparsers = parser.add_subparsers(dest='command', help='Available commands')
    
    # Display command
    display_parser = subparsers.add_parser('display', 
                                         prog='prodir display',  # Set display command name
                                         help='Display directory structure')
    display_parser.add_argument(
        'path',
        nargs='?',
        default=os.getcwd(),
        help='Path to analyze (default: current directory)'
    )
    display_parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='Show more detailed output'
    )
    
    # Create command
    create_parser = subparsers.add_parser('create',
                                        prog='prodir create',  # Set create command name
                                        help='Create directory structure')
    create_parser.add_argument(
        'file',
        help='File containing the directory structure'
    )
    create_parser.add_argument(
        '-o', '--output',
        default=os.getcwd(),
        help='Output directory (default: current directory)'
    )
    create_parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='Show more detailed output'
    )
    
    # Check if a direct path was provided
    if len(sys.argv) > 1 and not sys.argv[1].startswith('-') and not sys.argv[1] in ['display', 'create']:
        # Convert to display command with path
        sys.argv.insert(1, 'display')
    
    args = parser.parse_args()
    
    # If no command is specified, use display with current directory
    if not args.command:
        args.command = 'display'
        args.path = os.getcwd()
        args.verbose = False
    
    try:
        create_tree_structurer()
        
        if args.command == 'display':
            try:
                structure = get_structure(args.path)
                print(f"Analyzing directory: {args.path}")
                for line in structure:
                    print(line)
            except FileNotFoundError:
                print("Error: Directory does not exist", file=sys.stderr)
                sys.exit(1)
            except Exception as e:
                print(f"Error: {str(e)}", file=sys.stderr)
                sys.exit(1)
                
        elif args.command == 'create':
            if args.verbose:
                print(f"Creating directory structure in: {args.output}")
                print(f"Using structure from file: {args.file}")

            # Check if the output path exists
            if not os.path.exists(args.output):
                print(f"Error: The specified output path '{args.output}' does not exist.", file=sys.stderr)
                exit(1)

            try:
                create_structure_from_file(args.file, args.output)
                if args.verbose:
                    print("Structure created successfully")
                    print("\nResulting structure:")
                    structure = get_structure(args.output)
                    for line in structure:
                        print(line)
            except FileNotFoundError as e:
                if 'structure file' in str(e):
                    print("Error: Unable to open structure file", file=sys.stderr)
                else:
                    print("Error: Directory does not exist", file=sys.stderr)
                sys.exit(1)
            except Exception as e:
                print(f"Error: {str(e)}", file=sys.stderr)
                sys.exit(1)
                    
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()
